use base64::engine::general_purpose::STANDARD as BASE64;
use base64::engine::Engine as _;
use protobuf::Enum;
use serde::{Deserialize, Serialize};
use thiserror::Error;

use spotifyio_protocol::authentication::AuthenticationType;

use crate::Error;

#[derive(Debug, Error)]
pub enum AuthenticationError {
    #[error("unknown authentication type {0}")]
    AuthType(u32),
    #[error("invalid key")]
    Key,
}

impl From<AuthenticationError> for Error {
    fn from(err: AuthenticationError) -> Self {
        Error::invalid_argument(err)
    }
}

/// The credentials are used to log into the Spotify API.
#[derive(Debug, Clone, Default, Serialize, Deserialize, PartialEq)]
pub struct AuthCredentials {
    pub user_id: Option<String>,

    #[serde(serialize_with = "serialize_protobuf_enum")]
    #[serde(deserialize_with = "deserialize_protobuf_enum")]
    pub auth_type: AuthenticationType,

    #[serde(alias = "encoded_auth_blob")]
    #[serde(serialize_with = "serialize_base64")]
    #[serde(deserialize_with = "deserialize_base64")]
    pub auth_data: Vec<u8>,
}

impl AuthCredentials {
    pub fn with_access_token(token: impl Into<String>) -> Self {
        Self {
            user_id: None,
            auth_type: AuthenticationType::AUTHENTICATION_SPOTIFY_TOKEN,
            auth_data: token.into().into_bytes(),
        }
    }
}

fn serialize_protobuf_enum<T, S>(v: &T, ser: S) -> Result<S::Ok, S::Error>
where
    T: Enum,
    S: serde::Serializer,
{
    serde::Serialize::serialize(&v.value(), ser)
}

fn deserialize_protobuf_enum<'de, T, D>(de: D) -> Result<T, D::Error>
where
    T: Enum,
    D: serde::Deserializer<'de>,
{
    let v: i32 = serde::Deserialize::deserialize(de)?;
    T::from_i32(v).ok_or_else(|| serde::de::Error::custom("Invalid enum value"))
}

fn serialize_base64<T, S>(v: &T, ser: S) -> Result<S::Ok, S::Error>
where
    T: AsRef<[u8]>,
    S: serde::Serializer,
{
    serde::Serialize::serialize(&BASE64.encode(v.as_ref()), ser)
}

fn deserialize_base64<'de, D>(de: D) -> Result<Vec<u8>, D::Error>
where
    D: serde::Deserializer<'de>,
{
    let v: String = serde::Deserialize::deserialize(de)?;
    BASE64
        .decode(v)
        .map_err(|e| serde::de::Error::custom(e.to_string()))
}
