use std::{
    collections::HashMap,
    fs::File,
    io::BufReader,
    path::{Path, PathBuf},
    sync::{Arc, RwLock},
};

use serde::{Deserialize, Serialize};
use time::OffsetDateTime;
use uuid::Uuid;

use crate::{authentication::AuthCredentials, spclient::Token, Error, Session, SessionConfig};

type CacheData = HashMap<String, CachedSession>;

#[derive(Clone)]
pub struct ApplicationCache {
    inner: Arc<CacheInner>,
}

struct CacheInner {
    path: PathBuf,
    data: RwLock<CacheData>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
struct CachedSession {
    device_id: String,
    email: Option<String>,
    country: Option<String>,
    is_premium: bool,
    credentials: AuthCredentials,
    auth_token: Option<Token>,
    client_token: Option<Token>,
    #[serde(with = "time::serde::rfc3339", default = "OffsetDateTime::now_utc")]
    created_at: OffsetDateTime,
}

impl Default for CachedSession {
    fn default() -> Self {
        Self {
            device_id: Uuid::new_v4().simple().to_string(),
            email: None,
            country: None,
            is_premium: false,
            credentials: AuthCredentials::default(),
            auth_token: None,
            client_token: None,
            created_at: OffsetDateTime::now_utc(),
        }
    }
}

pub struct SessionCache {
    inner: Arc<CacheInner>,
    /// Username as cache key
    user_id: String,
}

pub struct AccountMetadata {
    pub user_id: String,
    pub email: Option<String>,
    pub created_at: OffsetDateTime,
}

impl ApplicationCache {
    pub fn new(path: PathBuf) -> Self {
        tracing::trace!("cache path: {path:?}");
        let data = if path.is_file() {
            match Self::read_file(&path) {
                Ok(data) => data,
                Err(e) => {
                    tracing::error!("could not read cache file: {e}");
                    CacheData::default()
                }
            }
        } else {
            CacheData::default()
        };

        Self {
            inner: CacheInner {
                path,
                data: RwLock::new(data),
            }
            .into(),
        }
    }

    pub fn new_session(&self, user_id: String) -> SessionCache {
        SessionCache {
            inner: self.inner.clone(),
            user_id,
        }
    }

    /// Get a list of device ids for all registered sessions
    pub fn device_ids(&self) -> Vec<String> {
        let data = self.inner.data.read().unwrap();
        data.keys().cloned().collect()
    }

    pub fn first_session(&self) -> Result<SessionCache, Error> {
        let data = self.inner.data.read().unwrap();
        if let Some(user_id) = data.keys().next() {
            tracing::debug!("first session: {user_id}");
            Ok(SessionCache {
                inner: self.inner.clone(),
                user_id: user_id.to_owned(),
            })
        } else {
            Err(Error::unauthenticated("no sessions available"))
        }
    }

    /// Get a list of sessions for all registered accounts
    pub fn sessions(&self, cfg: &SessionConfig, max_sessions: usize) -> Vec<Session> {
        let data = self.inner.data.read().unwrap();
        data.keys()
            .take(max_sessions)
            .map(|user_id| {
                let cache = SessionCache {
                    inner: self.inner.clone(),
                    user_id: user_id.to_owned(),
                };
                Session::new(cfg.clone(), cache)
            })
            .collect()
    }

    /// Get a list of registered Spotify accounts
    pub fn accounts(&self) -> Vec<AccountMetadata> {
        let data = self.inner.data.read().unwrap();
        let mut accounts = data
            .values()
            .filter_map(|session| {
                session
                    .credentials
                    .user_id
                    .as_deref()
                    .map(|user_id| AccountMetadata {
                        user_id: user_id.to_owned(),
                        email: session.email.clone(),
                        created_at: session.created_at,
                    })
            })
            .collect::<Vec<_>>();
        accounts.sort_by_key(|a| a.created_at);
        accounts
    }

    pub fn remove_account(&self, user_id: &str) -> Result<(), Error> {
        let mut data = self.inner.data.write().unwrap();
        if data.remove(user_id).is_none() {
            Err(Error::not_found("account does not exist"))
        } else {
            drop(data);
            self.inner.write_file();
            Ok(())
        }
    }

    fn read_file(path: &Path) -> Result<CacheData, Error> {
        let file = File::open(path)?;
        let data: CacheData = serde_json::from_reader(BufReader::new(file))?;
        Ok(data)
    }
}

impl CacheInner {
    fn _write_file(&self) -> Result<(), Error> {
        let entry = {
            let x = self.data.read().unwrap();
            x.clone()
        };
        let file = File::create(&self.path)?;
        serde_json::to_writer_pretty(file, &entry)?;
        Ok(())
    }

    fn write_file(&self) {
        if let Err(e) = self._write_file() {
            tracing::error!("could not write cache file: {e}");
        }
    }
}

impl SessionCache {
    #[cfg(test)]
    pub(crate) fn testing() -> Self {
        use path_macro::path;

        let cache = ApplicationCache::new(path!(
            env!("CARGO_MANIFEST_DIR") / ".." / ".." / "data" / "spotifyio.json"
        ));
        cache.first_session().unwrap()
    }

    pub fn read_device_id(&self) -> String {
        let data = self.inner.data.read().unwrap();
        if let Some(session) = data.get(&self.user_id) {
            return session.device_id.to_owned();
        }

        let mut data = self.inner.data.write().unwrap();
        let entry = data.entry(self.user_id.to_owned()).or_default();
        entry.device_id.to_owned()
    }

    pub fn read_email(&self) -> Option<String> {
        let data = self.inner.data.read().unwrap();
        data.get(&self.user_id).and_then(|d| d.email.to_owned())
    }

    pub fn read_is_premium(&self) -> bool {
        let data = self.inner.data.read().unwrap();
        data.get(&self.user_id)
            .map(|d| d.is_premium)
            .unwrap_or_default()
    }

    pub fn read_country(&self) -> Option<String> {
        let data = self.inner.data.read().unwrap();
        data.get(&self.user_id).and_then(|d| d.country.to_owned())
    }

    pub fn read_credentials(&self) -> Option<AuthCredentials> {
        let data = self.inner.data.read().unwrap();
        data.get(&self.user_id).map(|d| d.credentials.clone())
    }

    pub fn read_auth_token(&self) -> Option<Token> {
        let data = self.inner.data.read().unwrap();
        data.get(&self.user_id).and_then(|d| {
            d.auth_token
                .as_ref()
                .filter(|token| !token.is_expired())
                .cloned()
        })
    }

    pub fn read_client_token(&self) -> Option<Token> {
        let data = self.inner.data.read().unwrap();
        data.get(&self.user_id).and_then(|d| {
            d.client_token
                .as_ref()
                .filter(|token| !token.is_expired())
                .cloned()
        })
    }

    pub fn write_email(&self, email: Option<String>) {
        let mut data = self.inner.data.write().unwrap();
        let entry = data.entry(self.user_id.to_owned()).or_default();

        if entry.email != email {
            entry.email = email;
            drop(data);
            self.inner.write_file();
        }
    }

    pub fn write_country(&self, country: String) {
        let mut data = self.inner.data.write().unwrap();
        let entry = data.entry(self.user_id.to_owned()).or_default();

        if entry.country.as_deref() != Some(&country) {
            entry.country = Some(country);
            drop(data);
            self.inner.write_file();
        }
    }

    pub fn write_is_premium(&self, is_premium: bool) {
        let mut data = self.inner.data.write().unwrap();
        let entry = data.entry(self.user_id.to_owned()).or_default();

        if entry.is_premium != is_premium {
            entry.is_premium = is_premium;
            drop(data);
            self.inner.write_file();
        }
    }

    pub fn write_credentials(&self, credentials: AuthCredentials) {
        let mut data = self.inner.data.write().unwrap();
        let entry = data.entry(self.user_id.to_owned()).or_default();

        if entry.credentials != credentials {
            entry.credentials = credentials;
            drop(data);
            self.inner.write_file();
        }
    }

    pub fn write_auth_token(&self, auth_token: Token) {
        let mut data = self.inner.data.write().unwrap();
        let entry = data.entry(self.user_id.to_owned()).or_default();
        let new_token = Some(auth_token);

        if entry.auth_token != new_token {
            entry.auth_token = new_token;
            drop(data);
            self.inner.write_file();
        }
    }

    pub fn write_client_token(&self, client_token: Token) {
        let mut data = self.inner.data.write().unwrap();
        let entry = data.entry(self.user_id.to_owned()).or_default();
        let new_token = Some(client_token);

        if entry.client_token != new_token {
            entry.client_token = new_token;
            drop(data);
            self.inner.write_file();
        }
    }
}

#[cfg(test)]
mod tests {
    use super::CachedSession;

    #[test]
    fn t1() {
        let session = CachedSession::default();
        println!("{}", serde_json::to_string_pretty(&session).unwrap());
    }
}
