#![doc = include_str!("../README.md")]
#![cfg_attr(docsrs, feature(doc_cfg))]
#![warn(clippy::todo, clippy::dbg_macro)]

#[macro_use]
mod component;

mod apresolve;
mod audio_key;
mod authentication;
mod cache;
mod cdn_url;
mod connection;
mod error;
mod normalisation;
mod pagination;
mod pool;
mod session;
mod spclient;

pub mod model;
pub mod util;

#[cfg(feature = "oauth")]
#[cfg_attr(docsrs, doc(cfg(feature = "oauth")))]
pub mod oauth;

use std::{num::NonZeroU32, path::PathBuf, sync::Arc, time::Duration};

use futures_util::{stream, StreamExt};

use audio_key::AudioKeyManager;
use pool::ClientPool;
use spclient::SpClient;

pub use audio_key::{AudioKey, AudioKeyError};
pub use authentication::AuthCredentials;
pub use cache::{ApplicationCache, SessionCache};
pub use cdn_url::{CdnUrl, CdnUrlError};
pub use error::{Error, ErrorKind};
pub use governor::{Jitter, Quota};
pub use normalisation::NormalisationData;
pub use pool::PoolError;
pub use session::{Session, SessionConfig};
pub use spclient::RequestStrategy;
pub use spotifyio_model::{
    AlbumType, FileId, IdError, IncludeExternal, Market, RecommendationsAttribute, SearchResult,
    SearchType, SpotifyId, SpotifyItemType,
};
pub use spotifyio_protocol::authentication::AuthenticationType;

/// Protobuf enums
pub mod pb {
    pub use spotifyio_protocol::canvaz_meta::Type as CanvazType;
    pub use spotifyio_protocol::extended_metadata::ExtensionType;
    pub use spotifyio_protocol::extension_kind::ExtensionKind;
    pub use spotifyio_protocol::metadata::{
        album::Type as AlbumType,
        artist_with_role::ArtistRole,
        audio_file::Format as AudioFileFormat,
        copyright::Type as CopyrightType,
        episode::EpisodeType,
        image::Size as ImageSize,
        restriction::{Catalogue as CatalogueRestriction, Type as TypeRestriction},
        show::{ConsumptionOrder, MediaType},
    };
    pub use spotifyio_protocol::playlist4_external::{
        op::Kind as OpKind, source_info::Client, GeoblockBlockingType, ItemAttributeKind,
        ListAttributeKind,
    };
    pub use spotifyio_protocol::playlist_permission::PermissionLevel;
}

#[derive(Clone)]
pub struct SpotifyIoPool {
    inner: Arc<SpotifyIoPoolInner>,
}

struct SpotifyIoPoolInner {
    app_cache: ApplicationCache,
    pool: ClientPool<Session>,
}

pub struct PoolConfig {
    pub max_sessions: usize,
    pub max_n_auth_errors: u32,
    pub quota_key: Quota,
    pub jitter: Option<Jitter>,
}

impl Default for PoolConfig {
    fn default() -> Self {
        Self {
            max_sessions: usize::MAX,
            max_n_auth_errors: 3,
            quota_key: Quota::with_period(Duration::from_secs(300))
                .unwrap()
                .allow_burst(NonZeroU32::new(5).unwrap()),
            jitter: None,
        }
    }
}

impl SpotifyIoPool {
    pub fn new<P: Into<PathBuf>>(
        cache_path: P,
        session_cfg: &SessionConfig,
        pool_cfg: &PoolConfig,
    ) -> Result<Self, Error> {
        Self::from_app_cache(
            ApplicationCache::new(cache_path.into()),
            session_cfg,
            pool_cfg,
        )
    }

    pub fn from_app_cache(
        app_cache: ApplicationCache,
        session_cfg: &SessionConfig,
        pool_cfg: &PoolConfig,
    ) -> Result<Self, Error> {
        let sessions = app_cache.sessions(session_cfg, pool_cfg.max_sessions);
        let max_n_auth_errors = pool_cfg.max_n_auth_errors;
        let pool = ClientPool::new(pool_cfg.quota_key, pool_cfg.jitter, sessions, move |s| {
            s.n_auth_errors() > max_n_auth_errors
        })?;

        Ok(Self {
            inner: SpotifyIoPoolInner { app_cache, pool }.into(),
        })
    }

    pub fn app_cache(&self) -> &ApplicationCache {
        &self.inner.app_cache
    }

    pub async fn connect(&self) -> Result<(), Error> {
        let res = stream::iter(self.inner.pool.iter())
            .map(|session| async move {
                match session.connect_stored_creds().await {
                    Ok(_) => Ok(()),
                    Err(e) => {
                        tracing::error!("session {}: could not connect: {}", session.user_id(), e);
                        Err(e)
                    }
                }
            })
            .buffer_unordered(4)
            .collect::<Vec<_>>()
            .await;

        if res.iter().all(Result::is_err) {
            res.into_iter()
                .next()
                .unwrap_or(Err(Error::invalid_argument("no connections")))
        } else {
            let (n_available, n_errored) = self.n_available_errored();
            tracing::info!("{n_available} accounts connected");
            if n_errored > 0 {
                tracing::warn!("{n_errored} accounts could not be connected");
            }
            Ok(())
        }
    }

    pub fn n_available(&self) -> usize {
        self.inner.pool.n_available()
    }

    pub fn n_available_errored(&self) -> (usize, usize) {
        self.inner.pool.n_available_errored()
    }

    pub fn session(&self) -> Result<&Session, PoolError> {
        self.inner.pool.random()
    }

    pub fn spclient(&self) -> Result<&SpClient, PoolError> {
        Ok(self.inner.pool.random()?.spclient())
    }

    pub async fn audio_key(&self) -> Result<&AudioKeyManager, PoolError> {
        Ok(self.inner.pool.get().await?.audio_key())
    }

    pub async fn audio_key_timeout(
        &self,
        timeout: Duration,
    ) -> Result<&AudioKeyManager, PoolError> {
        Ok(self.inner.pool.get_timeout(timeout).await?.audio_key())
    }

    pub fn shutdown(&self) {
        for session in self.inner.pool.iter() {
            session.shutdown();
        }
    }
}
