use std::io::{Read, Seek, SeekFrom};

use byteorder::{LittleEndian, ReadBytesExt};

use crate::Error;

/// Audio metadata for volume normalization
///
/// Spotify provides these as `f32`, but audio metadata can contain up to `f64`.
/// Also, this negates the need for casting during sample processing.
///
/// More information about ReplayGain and how to apply this infomation to audio files
/// can be found here: <https://wiki.hydrogenaud.io/index.php?title=ReplayGain_1.0_specification>.
#[derive(Clone, Copy, Debug)]
pub struct NormalisationData {
    pub track_gain_db: f64,
    pub track_peak: f64,
    pub album_gain_db: f64,
    pub album_peak: f64,
}

impl NormalisationData {
    pub fn parse_from_ogg<T: Read + Seek>(mut file: T) -> Result<Self, Error> {
        const SPOTIFY_NORMALIZATION_HEADER_START_OFFSET: u64 = 144;
        let newpos = file.seek(SeekFrom::Start(SPOTIFY_NORMALIZATION_HEADER_START_OFFSET))?;
        if newpos != SPOTIFY_NORMALIZATION_HEADER_START_OFFSET {
            return Err(Error::failed_precondition(format!(
                "NormalisationData::parse_from_file seeking to {} but position is now {}",
                SPOTIFY_NORMALIZATION_HEADER_START_OFFSET, newpos
            )));
        }

        let track_gain_db = file.read_f32::<LittleEndian>()? as f64;
        let track_peak = file.read_f32::<LittleEndian>()? as f64;
        let album_gain_db = file.read_f32::<LittleEndian>()? as f64;
        let album_peak = file.read_f32::<LittleEndian>()? as f64;

        Ok(Self {
            track_gain_db,
            track_peak,
            album_gain_db,
            album_peak,
        })
    }
}
