use std::time::Duration;

use governor::{clock::Clock, DefaultDirectRateLimiter, Jitter, Quota};

use parking_lot::Mutex;
use rand::Rng;

#[derive(Debug, thiserror::Error)]
pub enum PoolError {
    #[error("no clients available")]
    Empty,
    #[error("timeout: {0:?}")]
    Timeout(Duration),
}

pub struct ClientPool<T> {
    robin: Mutex<usize>,
    entries: Vec<ClientPoolEntry<T>>,
    is_invalid: Box<dyn Fn(&T) -> bool>,
    jitter: Jitter,
}

struct ClientPoolEntry<T> {
    entry: T,
    limiter: DefaultDirectRateLimiter,
}

impl<T> ClientPool<T> {
    pub fn new<E: IntoIterator<Item = T>, F: Fn(&T) -> bool + 'static>(
        quota: Quota,
        jitter: Option<Jitter>,
        entries: E,
        is_invalid: F,
    ) -> Result<Self, PoolError> {
        let entries = entries
            .into_iter()
            .map(|entry| ClientPoolEntry {
                entry,
                limiter: DefaultDirectRateLimiter::direct(quota),
            })
            .collect::<Vec<_>>();
        if entries.is_empty() {
            return Err(PoolError::Empty);
        }

        Ok(Self {
            robin: Mutex::default(),
            entries,
            is_invalid: Box::new(is_invalid),
            jitter: jitter.unwrap_or_else(|| Jitter::up_to(quota.replenish_interval() / 2)),
        })
    }

    pub fn len(&self) -> usize {
        self.entries.len()
    }

    pub fn iter(&self) -> impl Iterator<Item = &T> {
        self.entries.iter().map(|e| &e.entry)
    }

    pub fn n_available(&self) -> usize {
        self.entries
            .iter()
            .filter(|entry| !(self.is_invalid)(&entry.entry))
            .count()
    }

    pub fn n_available_errored(&self) -> (usize, usize) {
        let n_available = self.n_available();
        (n_available, self.entries.len() - n_available)
    }

    fn wrapping_inc(&self, mut a: usize) -> usize {
        a = a.wrapping_add(1);
        if a >= self.len() {
            a = 0;
        }
        a
    }

    fn next_pos(&self, timeout: Option<Duration>) -> Result<(usize, Duration), PoolError> {
        let mut robin = self.robin.lock();
        let mut i = *robin;
        let mut max_to = Duration::ZERO;

        for _ in 0..((self.len() - 1).max(1)) {
            let entry = &self.entries[i];

            if !(self.is_invalid)(&entry.entry) {
                match entry.limiter.check() {
                    Ok(_) => {
                        *robin = self.wrapping_inc(i);
                        return Ok((i, Duration::ZERO));
                    }
                    Err(not_until) => {
                        let wait_time = not_until.wait_time_from(entry.limiter.clock().now());
                        if timeout.is_some_and(|to| wait_time > to) {
                            max_to = max_to.max(wait_time);
                        } else {
                            *robin = self.wrapping_inc(i);
                            return Ok((i, wait_time));
                        }
                    }
                }
            }
            i = self.wrapping_inc(i);
        }
        // If the pool is empty or all entries were skipped
        Err(if max_to == Duration::ZERO {
            PoolError::Empty
        } else {
            PoolError::Timeout(max_to)
        })
    }

    /// Get a client from the pool, waiting until the rate limit passes
    pub async fn get(&self) -> Result<&T, PoolError> {
        self._get_timeout(None).await
    }

    /// Get a client from the pool, waiting until the rate limit passes
    ///
    /// If the timeout is longer than the given value, the function returns a [`PoolError::Timeout`].
    pub async fn get_timeout(&self, timeout: Duration) -> Result<&T, PoolError> {
        self._get_timeout(Some(timeout)).await
    }

    /// Get a random, valid client from the pool, skipping the rate limit
    ///
    /// If no valid clients were found, [`PoolError::Empty`] is returned
    pub fn random(&self) -> Result<&T, PoolError> {
        let mut pos = rand::thread_rng().gen_range(0..self.len());

        for _ in 0..((self.len() - 1).max(1)) {
            let entry = &self.entries[pos].entry;
            if !(self.is_invalid)(entry) {
                return Ok(entry);
            }
            pos += 1;
        }
        Err(PoolError::Empty)
    }

    async fn _get_timeout(&self, timeout: Option<Duration>) -> Result<&T, PoolError> {
        let (i, wait_time) = self.next_pos(timeout)?;
        let entry = &self.entries[i];
        if !wait_time.is_zero() {
            let wait_with_jitter = self.jitter + wait_time;
            tracing::debug!("pool: waiting for {wait_with_jitter:?}");
            tokio::time::sleep(wait_with_jitter).await;
        }
        Ok(&entry.entry)
    }
}

#[cfg(test)]
mod tests {
    use std::{num::NonZeroU32, time::Duration};

    use tracing_test::traced_test;

    use super::*;

    const TIMEOUT: Duration = Duration::from_millis(100);

    #[tokio::test]
    #[traced_test]
    async fn one_client() {
        let pool = ClientPool::new(
            Quota::per_second(NonZeroU32::new(2).unwrap()),
            Some(Jitter::default()),
            [1],
            |_| false,
        )
        .unwrap();

        // Deplete the limiter
        for _ in 0..2 {
            assert_eq!(
                *tokio::time::timeout(TIMEOUT, pool.get())
                    .await
                    .unwrap()
                    .unwrap(),
                1
            );
        }
        assert!(tokio::time::timeout(TIMEOUT, pool.get()).await.is_err());
        if let PoolError::Timeout(to) = pool.get_timeout(TIMEOUT).await.unwrap_err() {
            assert!(to > TIMEOUT, "got {to:?}");
            tokio::time::sleep(to).await;
        } else {
            panic!("no timeout err");
        }

        assert_eq!(
            *tokio::time::timeout(TIMEOUT, pool.get())
                .await
                .unwrap()
                .unwrap(),
            1
        );

        assert_eq!(*pool.random().unwrap(), 1);
    }

    #[tokio::test]
    #[traced_test]
    async fn one_client_invalid() {
        let pool = ClientPool::new(
            Quota::per_minute(NonZeroU32::new(2).unwrap()),
            Some(Jitter::default()),
            [1],
            |_| true,
        )
        .unwrap();

        assert!(matches!(pool.get().await.unwrap_err(), PoolError::Empty));
        assert!(matches!(
            pool.get_timeout(TIMEOUT).await.unwrap_err(),
            PoolError::Empty
        ));
        assert!(matches!(pool.random().unwrap_err(), PoolError::Empty));
    }

    #[tokio::test]
    #[traced_test]
    async fn multiple_clients() {
        let pool = ClientPool::new(
            Quota::per_second(NonZeroU32::new(2).unwrap()),
            Some(Jitter::default()),
            [1, 2, 3, 4, 5], // active: 1, 3, 5
            |n| *n % 2 == 0,
        )
        .unwrap();

        let is_valid = |n: &i32| assert!(matches!(*n, 1 | 3 | 5));

        // Deplete the limiter
        for _ in 0..6 {
            is_valid(
                tokio::time::timeout(TIMEOUT, pool.get())
                    .await
                    .unwrap()
                    .unwrap(),
            );
        }

        assert!(tokio::time::timeout(TIMEOUT, pool.get()).await.is_err());
        if let PoolError::Timeout(to) = pool.get_timeout(TIMEOUT).await.unwrap_err() {
            assert!(to > TIMEOUT, "got {to:?}");
            tokio::time::sleep(to).await;
        } else {
            panic!("no timeout err");
        }

        is_valid(
            tokio::time::timeout(TIMEOUT, pool.get())
                .await
                .unwrap()
                .unwrap(),
        );

        is_valid(pool.random().unwrap());
    }
}
