use std::{
    collections::HashMap,
    future::Future,
    pin::Pin,
    sync::{Arc, Weak},
    task::{Context, Poll},
    time::{Duration, SystemTime, UNIX_EPOCH},
};

use byteorder::{BigEndian, ByteOrder};
use bytes::Bytes;
use futures_util::{StreamExt, TryStream};
use num_traits::FromPrimitive;
use once_cell::sync::OnceCell;
use parking_lot::RwLock;
use pin_project_lite::pin_project;
use quick_xml::events::Event;
use reqwest::Client;
use spotifyio_protocol::{authentication::AuthenticationType, keyexchange::ErrorCode};
use tokio::{
    select,
    sync::{mpsc, Notify},
    time::{Instant as TokioInstant, Sleep},
};
use tokio_stream::wrappers::UnboundedReceiverStream;
use tokio_util::sync::{CancellationToken, WaitForCancellationFutureOwned};
use tracing::{debug, error, info, trace, warn};
use url::Url;

use crate::{
    apresolve::ApResolver,
    audio_key::AudioKeyManager,
    authentication::AuthCredentials,
    cache::SessionCache,
    connection::{self, AuthenticationError, PacketType, Transport},
    error::Error,
    spclient::{RequestStrategy, SpClient},
    util::{self, SocketAddress},
    FileId,
};

#[derive(Debug, thiserror::Error)]
pub enum SessionError {
    #[error(transparent)]
    AuthenticationError(#[from] AuthenticationError),
    #[error("Cannot create session: {0}")]
    IoError(#[from] std::io::Error),
    #[error("Session is not connected")]
    NotConnected,
    #[error("packet {0} unknown")]
    Packet(u8),
}

impl From<SessionError> for Error {
    fn from(err: SessionError) -> Self {
        match err {
            SessionError::AuthenticationError(_) => Error::unauthenticated(err),
            SessionError::IoError(_) => Error::unavailable(err),
            SessionError::NotConnected => Error::unavailable(err),
            SessionError::Packet(_) => Error::unimplemented(err),
        }
    }
}

#[derive(Debug, Clone, Default)]
struct SessionData {
    auth_data: Vec<u8>,
    time_delta: i64,
    user_id: String,
    n_disconnects: u32,
    n_auth_errors: u32,
}

impl SessionData {
    fn from_creds(creds: Option<AuthCredentials>) -> Self {
        if let Some(creds) = creds {
            Self {
                user_id: creds.user_id.unwrap_or_default(),
                auth_data: creds.auth_data,
                ..Default::default()
            }
        } else {
            Self::default()
        }
    }
}

#[derive(Clone, Debug)]
pub struct SessionConfig {
    pub client_id: String,
    pub language: String,
    pub proxy: Option<Url>,
    pub ap_port: Option<u16>,
    pub request_strategy: RequestStrategy,
    pub pagination_chunks: u32,
}

impl Default for SessionConfig {
    fn default() -> Self {
        Self {
            client_id: util::SPOTIFY_CLIENT_ID.to_owned(),
            language: "en".to_owned(),
            proxy: None,
            ap_port: None,
            request_strategy: RequestStrategy::TryTimes(10),
            pagination_chunks: 50,
        }
    }
}

struct SessionInternal {
    config: SessionConfig,
    data: RwLock<SessionData>,
    cache: SessionCache,

    http_client: Client,
    init_lock: tokio::sync::Mutex<()>,
    connection: RwLock<Option<ConnectionData>>,

    apresolver: OnceCell<ApResolver>,
    audio_key: OnceCell<AudioKeyManager>,
    spclient: OnceCell<SpClient>,
}

struct ConnectionData {
    tx_connection: mpsc::UnboundedSender<(u8, Vec<u8>)>,
    cancel_connection: CancellationToken,
}

#[derive(Clone)]
pub struct Session(Arc<SessionInternal>);

impl Session {
    pub fn new(config: SessionConfig, cache: SessionCache) -> Self {
        Self(Arc::new(SessionInternal {
            config,
            data: RwLock::new(SessionData::from_creds(cache.read_credentials())),
            cache,
            http_client: Client::builder()
                .user_agent(util::USER_AGENT)
                .build()
                .unwrap(),
            init_lock: tokio::sync::Mutex::new(()),
            connection: RwLock::new(None),
            apresolver: OnceCell::new(),
            audio_key: OnceCell::new(),
            spclient: OnceCell::new(),
        }))
    }

    async fn connect_inner(
        &self,
        access_point: &SocketAddress,
        credentials: AuthCredentials,
    ) -> Result<(AuthCredentials, Transport), Error> {
        const MAX_RETRIES: u8 = 1;
        let mut transport = connection::connect_with_retry(
            &access_point.0,
            access_point.1,
            self.config().proxy.as_ref(),
            MAX_RETRIES,
        )
        .await?;
        let mut reusable_credentials = connection::authenticate(
            &mut transport,
            credentials.clone(),
            &self.cache().read_device_id(),
        )
        .await?;

        // Might be able to remove this once keymaster is replaced with login5.
        if credentials.auth_type == AuthenticationType::AUTHENTICATION_SPOTIFY_TOKEN {
            trace!(
                "Reconnect using stored credentials as token authed sessions cannot use keymaster."
            );
            transport = connection::connect_with_retry(
                &access_point.0,
                access_point.1,
                self.config().proxy.as_ref(),
                MAX_RETRIES,
            )
            .await?;
            reusable_credentials = connection::authenticate(
                &mut transport,
                reusable_credentials.clone(),
                &self.cache().read_device_id(),
            )
            .await?;
        }

        Ok((reusable_credentials, transport))
    }

    pub async fn connect_stored_creds(&self) -> Result<(), Error> {
        if self.is_connected() {
            return Ok(());
        }

        if let Some(creds) = self.cache().read_credentials() {
            self.connect(creds).await
        } else {
            Err(Error::unauthenticated("no stored credentials"))
        }
    }

    pub async fn connect(&self, credentials: AuthCredentials) -> Result<(), Error> {
        // Prevent multiple concurrent connection attempts
        let _init_lock = self.0.init_lock.lock().await;
        if self.is_connected() {
            return Ok(());
        }

        // There currently happen to be 6 APs but anything will do to avoid an infinite loop.
        const MAX_AP_TRIES: u8 = 6;
        let mut num_ap_tries = 0;
        let (reusable_credentials, transport) = loop {
            let ap = self.apresolver().resolve("accesspoint").await?;
            info!("Connecting to AP \"{}:{}\"", ap.0, ap.1);
            match self.connect_inner(&ap, credentials.clone()).await {
                Ok(ct) => break ct,
                Err(e) => {
                    num_ap_tries += 1;
                    if MAX_AP_TRIES == num_ap_tries {
                        error!("Tried too many access points");
                        return Err(e);
                    }
                    if let Some(AuthenticationError::LoginFailed(ErrorCode::TryAnotherAP)) =
                        e.error.downcast_ref::<AuthenticationError>()
                    {
                        warn!("Instructed to try another access point...");
                        continue;
                    } else if let Some(AuthenticationError::LoginFailed(..)) =
                        e.error.downcast_ref::<AuthenticationError>()
                    {
                        self.inc_auth_errors();
                        return Err(e);
                    } else {
                        warn!("Try another access point...");
                        continue;
                    }
                }
            }
        };

        let user_id = reusable_credentials
            .user_id
            .as_ref()
            .map_or("UNKNOWN", |s| s.as_str());
        info!("Authenticated as '{user_id}' !");
        self.set_user_id(user_id);
        self.set_auth_data(&reusable_credentials.auth_data);
        self.reset_auth_errors();

        self.cache().write_credentials(reusable_credentials);

        // This channel serves as a buffer for packets and serializes access to the TcpStream, such
        // that `self.send_packet` can return immediately and needs no additional synchronization.
        let (tx_connection, rx_connection) = mpsc::unbounded_channel();
        let cancel_conn = CancellationToken::new();

        {
            let mut conn_data = self.0.connection.write();
            *conn_data = Some(ConnectionData {
                tx_connection,
                cancel_connection: cancel_conn.clone(),
            });
        }

        let (sink, stream) = transport.split();
        let sender_task = UnboundedReceiverStream::new(rx_connection)
            .map(Ok)
            .forward(sink);
        let session_weak = self.weak();
        let cancel_conn2 = cancel_conn.clone();
        tokio::spawn(async move {
            select! {
                _ = cancel_conn2.cancelled() => {
                    trace!("sender_task cancelled");
                }
                x = sender_task => {
                    if let Err(e) = x {
                        error!("{}", e);
                        if let Some(session) = session_weak.try_upgrade() {
                            session.shutdown();
                        }
                    }
                }
            }
        });

        let task = DispatchTask::new(self.weak(), stream, cancel_conn);
        let initialized = task.initialized.clone();
        tokio::spawn(task);

        if tokio::time::timeout(Duration::from_secs(5), initialized.notified())
            .await
            .is_err()
        {
            return Err(Error::deadline_exceeded(
                "connection could not be initialized",
            ));
        }

        Ok(())
    }

    pub fn cache(&self) -> &SessionCache {
        &self.0.cache
    }

    pub fn apresolver(&self) -> &ApResolver {
        self.0
            .apresolver
            .get_or_init(|| ApResolver::new(self.weak()))
    }

    pub fn audio_key(&self) -> &AudioKeyManager {
        self.0
            .audio_key
            .get_or_init(|| AudioKeyManager::new(self.weak()))
    }

    pub fn http_client(&self) -> &Client {
        &self.0.http_client
    }

    pub fn spclient(&self) -> &SpClient {
        self.0.spclient.get_or_init(|| SpClient::new(self.weak()))
    }

    pub fn send_packet(&self, cmd: PacketType, data: Vec<u8>) -> Result<(), Error> {
        match self.0.connection.read().as_ref() {
            Some(conn) => Ok(conn.tx_connection.send((cmd as u8, data))?),
            None => Err(SessionError::NotConnected.into()),
        }
    }

    pub fn config(&self) -> &SessionConfig {
        &self.0.config
    }

    fn weak(&self) -> SessionWeak {
        SessionWeak(Arc::downgrade(&self.0))
    }

    pub fn device_id(&self) -> String {
        self.cache().read_device_id()
    }

    pub fn client_id(&self) -> &str {
        &self.0.config.client_id
    }

    pub fn auth_data(&self) -> Vec<u8> {
        self.0.data.read().auth_data.clone()
    }

    fn set_auth_data(&self, auth_data: &[u8]) {
        self.0.data.write().auth_data = auth_data.to_owned();
    }

    pub fn user_id(&self) -> String {
        self.0.data.read().user_id.clone()
    }

    fn set_user_id(&self, user_id: &str) {
        user_id.clone_into(&mut self.0.data.write().user_id);
    }

    pub fn email(&self) -> Option<String> {
        self.cache().read_email()
    }

    pub fn country(&self) -> Option<String> {
        self.cache().read_country()
    }

    pub fn is_premium(&self) -> bool {
        self.cache().read_is_premium()
    }

    pub fn catalogue(&self) -> &'static str {
        if self.cache().read_is_premium() {
            "premium"
        } else {
            "free"
        }
    }

    pub fn image_url(&self, image_id: &FileId) -> Result<String, Error> {
        Ok(format!(
            "https://i.scdn.co/image/{}",
            &image_id.to_base16()?
        ))
    }

    pub fn audio_preview_url(&self, preview_id: &FileId) -> Result<String, Error> {
        Ok(format!(
            "https://p.scdn.co/mp3-preview/{}",
            &preview_id.to_base16()?
        ))
    }

    pub fn shutdown(&self) {
        debug!("Shutdown: Invalidating session");
        {
            let mut dt = self.0.data.write();
            dt.n_disconnects += 1;
        }
        let mut conn = self.0.connection.write();
        if let Some(conn) = conn.as_ref() {
            conn.cancel_connection.cancel();
        }
        *conn = None;
    }

    pub fn is_connected(&self) -> bool {
        self.0.connection.read().is_some()
    }

    /// How often the session disconnected from Spotify.
    ///
    /// The client disconnects if it does not receive any Pings from Spotify or
    /// if the TCP connection is closed unexpectedly.
    pub fn n_disconnects(&self) -> u32 {
        self.0.data.read().n_disconnects
    }

    /// How often the session failed to authenticate with Spotify.
    ///
    /// The counter is reset on every successful connection.
    pub fn n_auth_errors(&self) -> u32 {
        self.0.data.read().n_auth_errors
    }

    pub(crate) fn inc_auth_errors(&self) {
        let mut data = self.0.data.write();
        data.n_auth_errors += 1;
    }

    pub(crate) fn reset_auth_errors(&self) {
        let mut data = self.0.data.write();
        data.n_auth_errors = 0;
    }
}

#[derive(Clone)]
pub struct SessionWeak(Weak<SessionInternal>);

impl SessionWeak {
    fn try_upgrade(&self) -> Option<Session> {
        self.0.upgrade().map(Session)
    }

    pub(crate) fn upgrade(&self) -> Session {
        self.try_upgrade()
            .expect("session was dropped and so should have this component")
    }
}

#[derive(Clone, Copy, Default, Debug, PartialEq)]
enum KeepAliveState {
    #[default]
    // Expecting a Ping from the server, either after startup or after a PongAck.
    ExpectingPing,

    // We need to send a Pong at the given time.
    PendingPong,

    // We just sent a Pong and wait for it be ACK'd.
    ExpectingPongAck,
}

const INITIAL_PING_TIMEOUT: Duration = Duration::from_secs(20);
const PING_TIMEOUT: Duration = Duration::from_secs(80); // 60s expected + 20s buffer
const PONG_DELAY: Duration = Duration::from_secs(60);
const PONG_ACK_TIMEOUT: Duration = Duration::from_secs(20);

impl KeepAliveState {
    fn debug(&self, sleep: &Sleep) {
        let delay = sleep
            .deadline()
            .checked_duration_since(TokioInstant::now())
            .map(|t| t.as_secs_f64())
            .unwrap_or(f64::INFINITY);

        trace!("keep-alive state: {:?}, timeout in {:.1}", self, delay);
    }
}

pin_project! {
    struct DispatchTask<S>
    where
        S: TryStream<Ok = (u8, Bytes)>
    {
        session: SessionWeak,
        keep_alive_state: KeepAliveState,
        #[pin]
        stream: S,
        #[pin]
        timeout: Sleep,
        initialized: Arc<Notify>,
        #[pin]
        cancelled: WaitForCancellationFutureOwned,
    }
}

impl<S> DispatchTask<S>
where
    S: TryStream<Ok = (u8, Bytes)>,
{
    fn new(session: SessionWeak, stream: S, cancel_conn: CancellationToken) -> Self {
        Self {
            session,
            keep_alive_state: KeepAliveState::ExpectingPing,
            stream,
            timeout: tokio::time::sleep(INITIAL_PING_TIMEOUT),
            initialized: Notify::new().into(),
            cancelled: cancel_conn.cancelled_owned(),
        }
    }

    fn dispatch(
        mut self: Pin<&mut Self>,
        session: &Session,
        cmd: u8,
        data: Bytes,
    ) -> Result<(), Error> {
        use KeepAliveState::*;
        use PacketType::*;

        let packet_type = FromPrimitive::from_u8(cmd);
        let cmd = match packet_type {
            Some(cmd) => cmd,
            None => {
                trace!("Ignoring unknown packet {:x}", cmd);
                return Err(SessionError::Packet(cmd).into());
            }
        };

        match packet_type {
            Some(Ping) => {
                trace!("Received Ping");
                if self.keep_alive_state != ExpectingPing {
                    warn!("Received unexpected Ping from server")
                }
                let mut this = self.as_mut().project();
                *this.keep_alive_state = PendingPong;
                this.timeout
                    .as_mut()
                    .reset(TokioInstant::now() + PONG_DELAY);
                this.keep_alive_state.debug(&this.timeout);

                let server_timestamp = BigEndian::read_u32(data.as_ref()) as i64;
                let timestamp = SystemTime::now()
                    .duration_since(UNIX_EPOCH)
                    .unwrap_or(Duration::ZERO)
                    .as_secs() as i64;
                {
                    let mut data = session.0.data.write();
                    data.time_delta = server_timestamp.saturating_sub(timestamp);
                }

                Ok(())
            }
            Some(PongAck) => {
                trace!("Received PongAck");
                if self.keep_alive_state != ExpectingPongAck {
                    warn!("Received unexpected PongAck from server")
                }
                let mut this = self.as_mut().project();
                *this.keep_alive_state = ExpectingPing;
                this.timeout
                    .as_mut()
                    .reset(TokioInstant::now() + PING_TIMEOUT);
                this.keep_alive_state.debug(&this.timeout);

                Ok(())
            }
            Some(CountryCode) => {
                let country = String::from_utf8(data.as_ref().to_owned())?;
                info!("Country: {:?}", country);
                session.cache().write_country(country);
                Ok(())
            }
            Some(AesKey) | Some(AesKeyError) => session.audio_key().dispatch(cmd, data),
            Some(ProductInfo) => {
                let data = std::str::from_utf8(&data)?;
                let mut reader = quick_xml::Reader::from_str(data);

                let mut buf = Vec::new();
                let mut current_element = String::new();
                let mut user_attributes = HashMap::new();

                loop {
                    match reader.read_event_into(&mut buf) {
                        Ok(Event::Start(ref element)) => {
                            std::str::from_utf8(element)?.clone_into(&mut current_element)
                        }
                        Ok(Event::End(_)) => {
                            current_element = String::new();
                        }
                        Ok(Event::Text(ref value)) => {
                            if !current_element.is_empty() {
                                let _ = user_attributes
                                    .insert(current_element.clone(), value.unescape()?.to_string());
                            }
                        }
                        Ok(Event::Eof) => break,
                        Ok(_) => (),
                        Err(e) => warn!(
                            "Error parsing XML at position {}: {:?}",
                            reader.buffer_position(),
                            e
                        ),
                    }
                }

                let is_premium = user_attributes.get("type").is_some_and(|t| t == "premium");
                session.cache().write_is_premium(is_premium);
                trace!("Received user data: is_premium={is_premium:?}");

                self.initialized.notify_waiters();

                Ok(())
            }
            Some(SecretBlock)
            | Some(LegacyWelcome)
            | Some(UnknownDataAllZeros)
            | Some(LicenseVersion)
            | Some(MercuryReq)
            | Some(MercurySub)
            | Some(MercuryUnsub)
            | Some(MercuryEvent)
            | Some(StreamChunkRes)
            | Some(ChannelError) => Ok(()),
            _ => {
                trace!("Ignoring {:?} packet with data {:#?}", cmd, data);
                Err(SessionError::Packet(cmd as u8).into())
            }
        }
    }
}

impl<S> Future for DispatchTask<S>
where
    S: TryStream<Ok = (u8, Bytes), Error = std::io::Error>,
    <S as TryStream>::Ok: std::fmt::Debug,
{
    type Output = Result<(), S::Error>;

    fn poll(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        use KeepAliveState::*;

        let session = match self.session.try_upgrade() {
            Some(session) => session,
            None => return Poll::Ready(Ok(())),
        };

        if self.as_mut().project().cancelled.poll(cx).is_ready() {
            trace!("dispatch_task cancelled");
            return Poll::Ready(Ok(()));
        }

        // Process all messages that are immediately ready
        loop {
            match self.as_mut().project().stream.try_poll_next(cx) {
                Poll::Ready(Some(Ok((cmd, data)))) => {
                    let result = self.as_mut().dispatch(&session, cmd, data);
                    if let Err(e) = result {
                        debug!("could not dispatch command: {}", e);
                    }
                }
                Poll::Ready(None) => {
                    warn!("[{}] Connection to server closed.", session.user_id());
                    session.shutdown();
                    return Poll::Ready(Ok(()));
                }
                Poll::Ready(Some(Err(e))) => {
                    warn!("[{}] Connection to server closed.", session.user_id());
                    session.shutdown();
                    return Poll::Ready(Err(e));
                }
                Poll::Pending => break,
            }
        }

        // Handle the keep-alive sequence, returning an error when we haven't received a
        // Ping/PongAck for too long.
        //
        // The expected keepalive sequence is
        // - Server: Ping
        // - wait 60s
        // - Client: Pong
        // - Server: PongAck
        // - wait 60s
        // - repeat
        //
        // This means that we silently lost connection to Spotify servers if
        // - we don't receive Ping immediately after connecting,
        // - we don't receive a Ping 60s after the last PongAck or
        // - we don't receive a PongAck immediately after our Pong.
        //
        // Currently, we add a safety margin of 20s to these expected deadlines.
        let mut this = self.as_mut().project();
        if let Poll::Ready(()) = this.timeout.as_mut().poll(cx) {
            match this.keep_alive_state {
                ExpectingPing | ExpectingPongAck => {
                    warn!(
                        "session lost connection to server ({:?})",
                        this.keep_alive_state
                    );
                    session.shutdown();
                    return Poll::Ready(Err(std::io::Error::new(
                        std::io::ErrorKind::TimedOut,
                        format!(
                            "session lost connection to server ({:?})",
                            this.keep_alive_state
                        ),
                    )));
                }
                PendingPong => {
                    trace!("Sending Pong");
                    // TODO: Ideally, this should flush the `Framed<TcpStream> as Sink`
                    // before starting the timeout.
                    let _ = session.send_packet(PacketType::Pong, vec![0, 0, 0, 0]);
                    *this.keep_alive_state = ExpectingPongAck;
                    this.timeout
                        .as_mut()
                        .reset(TokioInstant::now() + PONG_ACK_TIMEOUT);
                    this.keep_alive_state.debug(&this.timeout);
                }
            }
        }
        Poll::Pending
    }
}

#[cfg(test)]
mod tests {
    use crate::{FileId, SpotifyId};

    use super::*;
    use data_encoding::{BASE64, HEXLOWER};
    use hex_lit::hex;
    use tracing_test::traced_test;

    #[tokio::test]
    #[traced_test]
    async fn test_connection() {
        let s = Session::new(SessionConfig::default(), SessionCache::testing());
        s.connect_stored_creds().await.unwrap();
        // let track_id = SpotifyId::from_uri("spotify:track:1t2qYCAjUAoGfeFeoBlK51").unwrap();
        // let file_id = FileId::from_base16("80c50c5c305f66562990e52705876c21bc263fd7").unwrap();

        let client_token = s.spclient().client_token().await.unwrap();
        let auth_token = s.spclient().auth_token().await.unwrap();
        dbg!(client_token, auth_token);
    }

    /*
    #[tokio::test]
    #[traced_test]
    async fn test_connection_cancel() {
        let s = Session::new(
            SessionConfig {
                no_connection: true,
                ..Default::default()
            },
            SessionCache::testing(),
        );
        s.connect_stored_creds().await.unwrap();
        // let track_id = SpotifyId::from_uri("spotify:track:1t2qYCAjUAoGfeFeoBlK51").unwrap();
        // let file_id = FileId::from_base16("80c50c5c305f66562990e52705876c21bc263fd7").unwrap();

        let client_token = s.spclient().client_token().await.unwrap();
        let auth_token = s.spclient().auth_token().await.unwrap();
        tokio::time::sleep(Duration::from_secs(1)).await;

        debug!("client_token: {}\nauth_token: {}", client_token, auth_token);
    }
    */

    /*
    #[tokio::test]
    #[traced_test]
    async fn playplay_key() {
        let s = Session::new(
            SessionConfig {
                client_id: "65b708073fc0480ea92a077233ca87bd".to_owned(),
                proxy: None,
                ap_port: None,
            },
            SessionCache::testing(),
        );
        s.connect_stored_creds().await.unwrap();

        let token: [u8; 16] = hex!("015c8801577620d7d46de9d696bb9574"); // old (broken)
        // let token: [u8; 16] = hex!("018703f7ecf6cefbc8a7c2aa9730ca55"); // current playplay
        // let token: [u8; 16] = hex!("012c86c326e9ea6900486490e8751ec4"); // android
        // let token: [u8; 16] = hex!("013dcdf5262cb54f5b52694979192ad8"); // android

        // let token: [u8; 16] = [0x1, 0x5c, 0x88, 0x1, 0x57, 0x76, 0x20, 0xd7, 0xd4, 0x6d, 0xe9, 0xd6, 0x96, 0xbb, 0x95, 0x74]; // suischan

        let file_id = FileId::from_base16("560ad6f0d543aa20e7851f891b4e40ba47d54b54").unwrap();
        let obfuscated_key = s
            .spclient()
            .get_obfuscated_key(&file_id, &token, false)
            .await
            .unwrap();
        // 3740c244f4a5553e9ed37ee5f0b7b411
        dbg!(HEXLOWER.encode(&obfuscated_key));
    }
    */

    #[tokio::test]
    #[traced_test]
    async fn connection_key() {
        let s = Session::new(SessionConfig::default(), SessionCache::testing());
        s.connect_stored_creds().await.unwrap();

        let spotify_id = SpotifyId::from_base62("4AY7RuEn7nmnxtvya5ygCt").unwrap();
        let file_id = FileId::from_base16("560ad6f0d543aa20e7851f891b4e40ba47d54b54").unwrap();
        let key = s.audio_key().request(&spotify_id, &file_id).await.unwrap();
        assert_eq!(key.0, hex!("0ec8107d77338dcb4d31d5feb4e220fb"));
    }

    #[test]
    fn playlist_rev() {
        // Revision from protobuf API
        let rev: [u8; 24] = [
            // Number
            0, 0, 0, 12, //
            // Revision
            224, 63, 251, 128, 52, 154, 139, 88, 248, 198, 81, 40, 105, 177, 11, 116, 51, 250, 91,
            156,
        ];

        // Revision from JSON API (base64)
        assert_eq!(BASE64.encode(&rev), "AAAADOA/+4A0motY+MZRKGmxC3Qz+luc");

        assert_eq!(BigEndian::read_u32(&rev[0..4]), 12);
        assert_eq!(
            HEXLOWER.encode(&rev[4..]),
            "e03ffb80349a8b58f8c6512869b10b7433fa5b9c"
        );
    }

    /*
    fn print_album_group(g: &[AlbumGroup]) {
        for group in g {
            for album in &group.album {
                let id = SpotifyId::from_raw(album.gid(), SpotifyItemType::Album).unwrap();
                print!("{}; ", id.to_base62().unwrap());
            }
            println!();
        }
    }
    */
}
